/*
   OpenCL Example Program that perform Matrix multiply of NxN matrices
   Developed as training practical for EEE4120F/EEE4084F course.

   Project developed by J. Taylor and S. Winberg

   Illustrates the steps of OpenCL programme development as explained
   in the lectures.
*/

#include <time.h>
#include "main.h"
//------------------------------------------------------------------------------

// GPU reset?  See <http://stackoverflow.com/questions/12259044/
//                  limitations-of-work-item-load-in-gpu-cuda-opencl>

// <https://social.technet.microsoft.com/Forums/windows/en-US/
// 92a45329-3dd1-4c42-8a53-42dd232edd81/
// how-to-turn-off-timeout-detection-and-recovery-of-gpus>
//------------------------------------------------------------------------------

/** This function does a complete Matrix multiply of an NxN matrix, calculating
    the output elements: Sum of A(k,i) x B(j,k) for k=0..N-1. */
void Process_Serial(){
 int i, j, k;
 float result;

 for(j = 0; j < N; j++){
  for(i = 0; i < N; i++){
   result = 0;
   for(k = 0; k < N; k++){
    result += A[N*i + k] * B[N*k + j];
   }
   Output_Serial[N*j + i] = result;
  }
 }
}
//------------------------------------------------------------------------------

/** Set up the data on the GPU and Run the kernel. */
void Process_OpenCL(){
 printf("\n");

 // Send N as argument 3. Note that arguments are indexed from 0
 OpenCL_ConstantInt(3, N);

 // Send over the buffers A and B to the GPU memory, note that OpenCL_WriteData
 OpenCL_WriteData(A_Buffer, N*N*sizeof(float), A);
 OpenCL_WriteData(B_Buffer, N*N*sizeof(float), B);

 // Execute the kernel function
 OpenCL_Run(N, LocalSize);

 // Read back the result from the GPU
 OpenCL_ReadData(OutputBuffer, N*N*sizeof(float), Output_OpenCL);
}
//------------------------------------------------------------------------------

/** Compare the Output_Serial and Output_OpenCL matrices and
    indicate if they are the same or not. If the matrices
    are not too big (row length less than 10) print
    all the matrices for visual inspection. */
void Compare(){
 int i, j;

 // compare the two result matrices
 int same = 1;
 for(j = 0; j < N; j++){
  for(i = 0; i < N; i++){
    if (Output_Serial[N*j + i] != Output_OpenCL[N*j + i])
       same=0;
  }
 }
 if (same) printf(" Result matrices are the same :)\n");
     else  printf(" Result matrices differ!!\n");

 if (N>10) {
    printf(" Matrices too big to display\n");
    return;
 }

 printf("A:\n");
 for(j = 0; j < N; j++){
  for(i = 0; i < N; i++){
   printf("\t%g", A[N*j + i]);
  }
  printf("\n");
 }

 printf("B:\n");
 for(j = 0; j < N; j++){
  for(i = 0; i < N; i++){
   printf("\t%g", B[N*j + i]);
  }
  printf("\n");
 }

 printf("Output_Serial:\n");
 for(j = 0; j < N; j++){
  for(i = 0; i < N; i++){
   printf("\t%g", Output_Serial[N*j + i]);
  }
  printf("\n");
 }

 printf("Output_OpenCL:\n");
 for(j = 0; j < N; j++){
  for(i = 0; i < N; i++){
   printf("\t%g", Output_OpenCL[N*j + i]);
  }
  printf("\n");
 }
}
//------------------------------------------------------------------------------

/** Fill an NxN matrix with random values ranging from 0 to 20 */
void Fill(float* A){
 int i, j;

 for(j = 0; j < N; j++){
  for(i = 0; i < N; i++){
   A[N*j + i] = rand() % 20;
  }
 }
}
//------------------------------------------------------------------------------

/** Entry point to the CPU application */
int main(){
 // Initialise OpenCL
 if(
  !OpenCL_Init("NVIDIA"                ) && // nVidia
  !OpenCL_Init("Advanced Micro Devices") && // AMD
  !OpenCL_Init(0                       )    // Default
 ){
  printf("Error: Cannot initialise OpenCL.\n");
  return 1;
 }

 // Load a kernel
 if(!OpenCL_LoadKernel("OpenCL/Kernel.cl", "Multiply")) return 1;

 // Choose the matrix size and how much memory it needs
 N = 3;
 size_t BufferSize = N*N*sizeof(float);

 // Allocate CPU RAM
 A = (float*)malloc(BufferSize);
 B = (float*)malloc(BufferSize);

 Output_Serial = (float*)malloc(BufferSize);
 Output_OpenCL = (float*)malloc(BufferSize);

 // Provide seed to random number generator
 srand(time(NULL));

 // Put data into the matrices
 Fill(A);
 Fill(B);

 // Decide how to split up the array of processors into a LocalSize[0]xLocalSize[1] matrix of processors
 OpenCL_PrepareLocalSize(N, LocalSize);

 // Allocate GPU RAM
 A_Buffer     = OpenCL_CreateBuffer(0, CL_MEM_READ_ONLY , BufferSize); // Create buffer linked to kernel function argument 0
 B_Buffer     = OpenCL_CreateBuffer(1, CL_MEM_READ_ONLY , BufferSize); // linked to kernel argument 1
 OutputBuffer = OpenCL_CreateBuffer(2, CL_MEM_WRITE_ONLY, BufferSize); // linked to kernel argument 2

 // Process the matrices on CPU
 tic();
 Process_Serial();
 printf("Serial: %lg ms\n", toc()/1e-3);

 // Process the matrices on GPU
 tic();
 Process_OpenCL();
 printf("\nOpenCL: %lg ms\n\n", toc()/1e-3);

 // Compare results -- will only display small matrices
 Compare();

 // Clean-up
 free(A);
 free(B);
 free(Output_Serial);
 free(Output_OpenCL);

 OpenCL_FreeBuffer(A_Buffer    );
 OpenCL_FreeBuffer(B_Buffer    );
 OpenCL_FreeBuffer(OutputBuffer);
 OpenCL_Destroy();

 return 0;
}
//------------------------------------------------------------------------------
